@@ -16,8 +16,6 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import print_function
-
 import atexit
 import functools
 import logging
@@ -120,7 +118,6 @@ class VcTreeStore(tree.DiffTreeStore):
         tree.DiffTreeStore.__init__(self, 1, [str] * 5)
 
     def get_file_path(self, it):
-        # Use instead of value_path; does not incorrectly decode
         return self.get_value(it, self.column_index(tree.COL_PATH, 0))
 
 
@@ -155,13 +152,6 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         "ignored": ("VcShowIgnored", Entry.is_ignored),
     }
 
-    file_encoding = sys.getfilesystemencoding()
-
-    @classmethod
-    def display_path(cls, bytes):
-        encodings = (cls.file_encoding,)
-        return misc.fallback_decode(bytes, encodings, lossy=True)
-
     def __init__(self):
         melddoc.MeldDoc.__init__(self)
         gnomeglade.Component.__init__(self, "vcview.ui", "vcview",
@@ -239,30 +229,36 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         vcs_model = self.combobox_vcs.get_model()
         vcs_model.clear()
 
-        # VC systems work at the directory level, so make sure we're checking
-        # for VC support there instead of on a specific file.
+        # VC systems can be executed at the directory level, so make sure
+        # we're checking for VC support there instead of
+        # on a specific file or on deleted/unexisting path inside vc
         location = os.path.abspath(location or ".")
-        if os.path.isfile(location):
-            location = os.path.dirname(location)
+        while not os.path.isdir(location):
+            parent_location = os.path.dirname(location)
+            if len(parent_location) >= len(location):
+                # no existing parent: for example unexisting drive on Windows
+                break
+            location = parent_location
+        else:
+            # existing parent directory was found
+            for avc in vc.get_vcs(location):
+                err_str = ''
+                vc_details = {'name': avc.NAME, 'cmd': avc.CMD}
 
-        for avc in vc.get_vcs(location):
-            err_str = ''
-            vc_details = {'name': avc.NAME, 'cmd': avc.CMD}
+                if not avc.is_installed():
+                    # Translators: This error message is shown when a version
+                    # control binary isn't installed.
+                    err_str = _("%(name)s (%(cmd)s not installed)")
+                elif not avc.valid_repo(location):
+                    # Translators: This error message is shown when a version
+                    # controlled repository is invalid.
+                    err_str = _("%(name)s (Invalid repository)")
 
-            if not avc.is_installed():
-                # Translators: This error message is shown when a version
-                # control binary isn't installed.
-                err_str = _("%(name)s (%(cmd)s not installed)")
-            elif not avc.valid_repo(location):
-                # Translators: This error message is shown when a version
-                # controlled repository is invalid.
-                err_str = _("%(name)s (Invalid repository)")
+                if err_str:
+                    vcs_model.append([err_str % vc_details, avc, False])
+                    continue
 
-            if err_str:
-                vcs_model.append([err_str % vc_details, avc, False])
-                continue
-
-            vcs_model.append([avc.NAME, avc(location), True])
+                vcs_model.append([avc.NAME, avc(location), True])
 
         valid_vcs = [(i, r[1].NAME) for i, r in enumerate(vcs_model) if r[2]]
         default_active = min(valid_vcs)[0] if valid_vcs else 0
@@ -331,15 +327,14 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         return recent.TYPE_VC, [self.location]
 
     def recompute_label(self):
-        location = self.display_path(self.location)
-        self.label_text = os.path.basename(location)
+        self.label_text = os.path.basename(self.location)
         # TRANSLATORS: This is the location of the directory being viewed
-        self.tooltip_text = _("%s: %s") % (_("Location"), location)
+        self.tooltip_text = _("%s: %s") % (_("Location"), self.location)
         self.label_changed()
 
     def _search_recursively_iter(self, iterstart):
         rootname = self.model.get_file_path(iterstart)
-        display_prefix = len(self.display_path(rootname)) + 1
+        display_prefix = len(rootname) + 1
         symlinks_followed = set()
         todo = [(self.model.get_path(iterstart), rootname)]
 
@@ -354,7 +349,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             todo.sort()
             treepath, path = todo.pop(0)
             it = self.model.get_iter(treepath)
-            yield _("Scanning %s") % self.display_path(path)[display_prefix:]
+            yield _("Scanning %s") % path[display_prefix:]
 
             entries = self.vc.get_entries(path)
             entries = [e for e in entries if any(f(e) for f in filters)]
@@ -405,7 +400,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         path = directory.get_path()
         self.set_location(path)
 
-    def on_delete_event(self, appquit=0):
+    def on_delete_event(self):
         self.scheduler.remove_all_tasks()
         self.emit('close', 0)
         return Gtk.ResponseType.OK
@@ -427,7 +422,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             self.emit("create-diff", [path], {})
             return
 
-        basename = self.display_path(os.path.basename(path))
+        basename = os.path.basename(path)
         meta = {
             'parent': self,
             'prompt_resolve': False,
